import 'dart:convert';
import 'dart:typed_data';
import 'package:dio/dio.dart';

/// 网络日志打印
class PQLogInterceptor extends Interceptor {
  PQLogInterceptor({
    this.request = true,
    this.requestHeader = true,
    this.requestBody = true,
    this.responseHeader = false,
    this.responseBody = true,
    this.error = true,
    this.ignoreUrls,
    this.ignoreHost,
  });

  /// Print request [Options]
  bool request;

  /// Print request header [Options.headers]
  bool requestHeader;

  /// Print request data [Options.data]
  bool requestBody;

  /// Print [Response.data]
  bool responseBody;

  /// Print [Response.headers]
  bool responseHeader;

  /// Print error message
  bool error;

  /// 忽略的列表
  List<String>? ignoreUrls;

  /// 忽略的域名
  String? ignoreHost;

  /// 是有有忽略host
  bool get hasIgnoreHost => ignoreHost?.isNotEmpty ?? false;

  /// 是否有忽略url
  bool get hasIgnoreUrls => ignoreUrls?.isNotEmpty ?? false;

  @override
  Future onRequest(
      RequestOptions options, RequestInterceptorHandler handler) async {
    final uri = options.uri;
    if (_checkIgnore(uri)) {
      handler.next(options);
      return;
    }
    var _networkLog = _NetworkLog("onRequest");
    _networkLog.printTopLine();
    _networkLog.printLog('*** Request ***');
    printKV('uri', options.uri, _networkLog);
    _networkLog.printMiddleLine();

    if (request) {
      printKV('method', options.method, _networkLog);
      printKV('responseType', options.responseType.toString(), _networkLog);
      printKV('followRedirects', options.followRedirects, _networkLog);
      printKV('connectTimeout', options.connectTimeout, _networkLog);
      printKV('receiveTimeout', options.receiveTimeout, _networkLog);
      printKV('extra', options.extra, _networkLog);
    }

    //请求 头部 打印
    if (requestHeader) {
      _networkLog.printMiddleLine();
      _networkLog.printLog('headers:');
      _networkLog.printLog(options.headers);
    }

    //请求 参数 打印
    if (requestBody) {
      _networkLog.printMiddleLine();
      _networkLog.printLog('data:');
      _networkLog.printLog(options.data);
    }
    _networkLog.printBottomLine();
    handler.next(options);
  }

  @override
  Future onError(DioError err, ErrorInterceptorHandler handler) async {
    final uri = err.requestOptions.uri;
    if (_checkIgnore(uri)) {
      handler.next(err);
      return;
    }
    var _networkLog = _NetworkLog("onResponse");
    if (error) {
      _networkLog.printTopLine();
      _networkLog.printLog('*** DioError ***:');
      _networkLog.printLog('uri: ${err.requestOptions.uri}');
      _networkLog.printLog('$err');
      if (err.response != null) {
        _printResponse(err.response!, _networkLog);
      }
      _networkLog.printBottomLine();
    }
    handler.next(err);
  }

  @override
  Future onResponse(
      Response response, ResponseInterceptorHandler handler) async {
    final uri = response.requestOptions.uri;
    if (_checkIgnore(uri)) {
      handler.next(response);
      return;
    }
    var _networkLog = _NetworkLog("onResponse");
    _networkLog.printTopLine();
    _networkLog.printLog('*** Response ***');
    _printResponse(response, _networkLog);
    _networkLog.printBottomLine();
    handler.next(response);
  }

  //收到打印
  void _printResponse(Response response, _NetworkLog _networkLog) {
    printKV('uri', response.requestOptions.uri, _networkLog);
    printKV('statusCode', response.statusCode ?? "null", _networkLog);
    if (responseHeader) {
      if (response.isRedirect == true) {
        printKV('redirect', response.realUri, _networkLog);
      }

        _networkLog.printMiddleLine();
        _networkLog.printLog('headers:');
        _networkLog.printLog(response.headers);

    }
    //请求 收到 打印
    if (responseBody) {
      _networkLog.printMiddleLine();
      _networkLog.printLog('Response data:');
      _networkLog.printLog(response.data);
    }
  }

  void printKV(String key, Object v, _NetworkLog _networkLog) {
    _networkLog.printLog('$key: $v');
  }

  /// 检查是否忽略
  bool _checkIgnore(Uri uri) {
    /// 没有host并且没有urls，直接返回false
    if (!hasIgnoreHost && !hasIgnoreUrls) return false;

    /// 只要满足两个条件的一个就进行忽略
    return (ignoreUrls?.contains(uri.toString()) ?? false) ||
        uri.host == ignoreHost;
  }
}

class _NetworkLog {
  static const topLeftCorner = '┌';
  static const bottomLeftCorner = '└';
  static const middleCorner = '├';
  static const verticalLine = '│';
  static const doubleDivider = '─';
  static const singleDivider = '┄';

  static String? _topOrBottomLine;
  static String? _middleLine;

  String _tag;

  _NetworkLog([this._tag = ""]);

  void printTopLine() {
    _createTopOrBottomLineIfNeed();
    print('$_tag$topLeftCorner$_topOrBottomLine');
  }

  void printMiddleLine() {
    _createMiddleLineIfNeed();
    print('$_tag$middleCorner$_middleLine');
  }

  void printBottomLine() {
    _createTopOrBottomLineIfNeed();
    print('$_tag$bottomLeftCorner$_topOrBottomLine');
  }

  void printLog(dynamic message) {
    var msg = stringifyMessage(message);
    var split = msg.split('\n');
    if (split.isNotEmpty) {
      for (var line in msg.split('\n')) {
        print('$_tag$verticalLine $line');
      }
    } else {
      print('$_tag$verticalLine $message');
    }
  }

  void _createMiddleLineIfNeed() {
    if (_middleLine == null) {
      var doubleDividerLine = StringBuffer();
      for (var i = 0; i < 120; i++) {
        doubleDividerLine.write(singleDivider);
      }
      _middleLine = '$doubleDividerLine';
    }
  }

  void _createTopOrBottomLineIfNeed() {
    if (_topOrBottomLine == null) {
      var doubleDividerLine = StringBuffer();
      for (var i = 0; i < 120; i++) {
        doubleDividerLine.write(doubleDivider);
      }
      _topOrBottomLine = '$doubleDividerLine';
    }
  }

  String stringifyMessage(dynamic message) {
    if (message is Map || message is Iterable) {
      if (message is Uint8List)
        return message.toString();
      var encoder = JsonEncoder.withIndent('  ');
      return encoder.convert(message);
    } else {
      try {
        if (message is String) {
          Map map = json.decode(message);
          return stringifyMessage(map);
        }
      } catch (e) {}
      return message.toString();
    }
  }
}
