import 'dart:convert';
import 'dart:io';
import 'package:dio/adapter.dart';
import 'package:dio/dio.dart';
import 'log_interceptor.dart';
import 'method.dart';

/// 数据解析
typedef PQNetworkDataDecoder<T> = T Function(dynamic data);

/// 使用之前一定、一定、一定要先调用config方法初始化dio
abstract class PQNetwork {
  static late Dio _dio;

  static Dio get dio => _dio;

  /// 是否开启日志打印
  static bool _enableLog = true;

  // ignore: unnecessary_getters_setters
  static bool get enableLog => _enableLog;

  // ignore: unnecessary_getters_setters
  static set enableLog(bool enable) {
    _enableLog = enable;
    if (enable) {
      _dio.interceptors.add(_logInterceptor);
    } else {
      _dio.interceptors.removeWhere((element) => element is PQLogInterceptor);
    }
  }

  /// 代理
  static String? _proxy;

  // ignore: unnecessary_getters_setters
  static String? get proxy => _proxy;

  // ignore: unnecessary_getters_setters
  static set proxy(String? proxy) {
    if (proxy?.isEmpty ?? true) return;
    _proxy = proxy;
    (dio.httpClientAdapter as DefaultHttpClientAdapter).onHttpClientCreate =
        (HttpClient client) {
      client.findProxy = (uri) {
        ///这里的 192.168.0.102:8888就是我们的代理服务地址
        final str = "PROXY $proxy";
        return str;
      };
      client.badCertificateCallback = (
        X509Certificate cert,
        String host,
        int port,
      ) {
        return true;
      };
    };
  }

  /// 日志打印者
  static PQLogInterceptor _logInterceptor = PQLogInterceptor();

  // ignore: unnecessary_getters_setters
  static PQLogInterceptor get logInterceptor => _logInterceptor;

  // ignore: unnecessary_getters_setters
  static set logInterceptor(PQLogInterceptor value) {
    _logInterceptor = value;
  }

  /// 配置Dio
  static void config(
    BaseOptions options, {
    PQLogInterceptor? logInterceptor,
    bool enableLog = true,
  }) {
    _dio = Dio(options);
    if (logInterceptor != null) {
      PQNetwork.logInterceptor = logInterceptor;
    }
    PQNetwork.enableLog = enableLog;
  }

  /// 设置cookie,
  /// 传入的是null就会移除header之中的Cookie，
  /// 不为空就是设值
  static setCookie(String? cookie) {
    if (cookie == null && dio.options.headers.containsKey("Cookie")) {
      dio.options.headers.remove("Cookie");
    } else {
      dio.options.headers.addAll({'Cookie': cookie});
    }
  }

  /// get
  static Future<Response<T>> get<T>(
    String path, {
    Map<String, dynamic>? queryParameters,
    CancelToken? cancelToken,
    Options? options,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
    PQNetworkDataDecoder<T>? decoder,
    bool mock = false,
    dynamic mockData,
    Duration mockDuration = const Duration(milliseconds: 500),
  }) async {
    return await _request(
      Method.get,
      path,
      queryParameters: queryParameters,
      cancelToken: cancelToken,
      options: options,
      onSendProgress: onSendProgress,
      onReceiveProgress: onReceiveProgress,
      decoder: decoder,
      mock: mock,
      mockData: mockData,
      mockDuration: mockDuration,
    );
  }

  static Future<Response> getBytes(
    String path, {
    Map<String, dynamic>? queryParameters,
    CancelToken? cancelToken,
    Options? options,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
    PQNetworkDataDecoder? decoder,
    bool mock = false,
    dynamic mockData,
    Duration mockDuration = const Duration(milliseconds: 500),
  }) async {
    return await _request(
      Method.get,
      path,
      queryParameters: queryParameters,
      cancelToken: cancelToken,
      options: _checkOptions(Method.get, options)
        ..responseType = ResponseType.bytes,
      onSendProgress: onSendProgress,
      onReceiveProgress: onReceiveProgress,
      decoder: decoder,
      mock: mock,
      mockData: mockData,
      mockDuration: mockDuration,
    );
  }

  /// post
  static Future<Response<T>> post<T>(
    String path, {
    data,
    Map<String, dynamic>? queryParameters,
    CancelToken? cancelToken,
    Options? options,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
    PQNetworkDataDecoder<T>? decoder,
    bool mock = false,
    dynamic mockData,
    Duration mockDuration = const Duration(milliseconds: 500),
  }) async {
    return await _request(
      Method.post,
      path,
      data: data,
      queryParameters: queryParameters,
      cancelToken: cancelToken,
      options: options,
      onSendProgress: onSendProgress,
      onReceiveProgress: onReceiveProgress,
      decoder: decoder,
      mock: mock,
      mockData: mockData,
      mockDuration: mockDuration,
    );
  }

  /// put
  static Future<Response<T>> put<T>(
    String path, {
    data,
    CancelToken? cancelToken,
    Options? options,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
    PQNetworkDataDecoder<T>? decoder,
    bool mock = false,
    dynamic mockData,
    Duration mockDuration = const Duration(milliseconds: 500),
  }) async {
    return await _request(
      Method.put,
      path,
      data: data,
      cancelToken: cancelToken,
      options: options,
      onSendProgress: onSendProgress,
      onReceiveProgress: onReceiveProgress,
      decoder: decoder,
      mock: mock,
      mockData: mockData,
      mockDuration: mockDuration,
    );
  }

  /// 删除
  static Future<Response<T>> delete<T>(
    String path, {
    data,
    Map<String, dynamic>? queryParameters,
    CancelToken? cancelToken,
    Options? options,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
    PQNetworkDataDecoder<T>? decoder,
    bool mock = false,
    dynamic mockData,
    Duration mockDuration = const Duration(milliseconds: 500),
  }) async {
    return await _request(
      Method.put,
      path,
      data: data,
      queryParameters: queryParameters,
      cancelToken: cancelToken,
      options: options,
      onSendProgress: onSendProgress,
      onReceiveProgress: onReceiveProgress,
      decoder: decoder,
      mock: mock,
      mockData: mockData,
      mockDuration: mockDuration,
    );
  }

  static Future<Response<T>> head<T>(
    String path, {
    data,
    Map<String, dynamic>? queryParameters,
    CancelToken? cancelToken,
    Options? options,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
    PQNetworkDataDecoder<T>? decoder,
    bool mock = false,
    dynamic mockData,
    Duration mockDuration = const Duration(milliseconds: 500),
  }) async {
    return await _request(
      Method.head,
      path,
      data: data,
      queryParameters: queryParameters,
      cancelToken: cancelToken,
      options: options,
      onSendProgress: onSendProgress,
      onReceiveProgress: onReceiveProgress,
      mock: mock,
      mockData: mockData,
      mockDuration: mockDuration,
    );
  }

  /// patch
  static Future<Response<T>> patch<T>(
    String path, {
    data,
    Map<String, dynamic>? queryParameters,
    CancelToken? cancelToken,
    Options? options,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
    PQNetworkDataDecoder<T>? decoder,
    bool mock = false,
    dynamic mockData,
    Duration mockDuration = const Duration(milliseconds: 500),
  }) async {
    return await _request(
      Method.patch,
      path,
      data: data,
      queryParameters: queryParameters,
      cancelToken: cancelToken,
      options: options,
      onSendProgress: onSendProgress,
      onReceiveProgress: onReceiveProgress,
      mock: mock,
      mockData: mockData,
      mockDuration: mockDuration,
    );
  }

  /// 下载文件
  static Future<Response> download(
    String urlPath,
    savePath, {
    ProgressCallback? onReceiveProgress,
    Map<String, dynamic>? queryParameters,
    CancelToken? cancelToken,
    bool deleteOnError = true,
    String lengthHeader = Headers.contentLengthHeader,
    data,
    Options? options,
  }) async {
    return dio.download(
      urlPath,
      savePath,
      onReceiveProgress: onReceiveProgress,
      queryParameters: queryParameters,
      cancelToken: cancelToken,
      deleteOnError: deleteOnError,
      lengthHeader: lengthHeader,
      data: data,
      options: options,
    );
  }

  /// 根据bytes进行上传
  static Future<Response<T>> upload<T>(
    String path,
    List<num> bytes, {
    Method method = Method.put,
    CancelToken? cancelToken,
    Options? options,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
  }) async {
    return await _request(
      method,
      path,
      data: Stream.fromIterable(bytes.map((e) => [e])),
      cancelToken: cancelToken,
      options: options,
      onSendProgress: onSendProgress,
      onReceiveProgress: onReceiveProgress,
    );
  }

  /// 根据文件路径上传
  static Future<Response<T>> uploadWithFilePath<T>(
    String path,
    String filePath, {
    Method method = Method.put,
    CancelToken? cancelToken,
    Options? options,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
  }) async {
    try {
      final bytes = File(filePath).readAsBytesSync();
      return await upload(
        path,
        bytes,
        method: method,
        cancelToken: cancelToken,
        options: options,
        onSendProgress: onSendProgress,
        onReceiveProgress: onReceiveProgress,
      );
    } catch (e) {
      RequestOptions requestOptions = RequestOptions(path: path);
      return Response(
          requestOptions: requestOptions,
          statusCode: 0,
          statusMessage: "ReadFile error ${e.toString()}");
    }
  }

  static Future<Response<T>> _request<T>(
    Method method,
    String path, {
    data,
    Map<String, dynamic>? queryParameters,
    CancelToken? cancelToken,
    Options? options,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
    PQNetworkDataDecoder<T>? decoder,
    bool mock = false,
    dynamic mockData,
    Duration mockDuration = const Duration(milliseconds: 500),
  }) async {
    if (mock) {
      return Future.delayed(
          mockDuration, () => _mock<T>(decoder, mockData, path, 'GET'));
    }
    try {
      final res = await dio.request(path,
          data: data,
          queryParameters: queryParameters,
          cancelToken: cancelToken,
          options: _checkOptions(method, options),
          onSendProgress: onSendProgress,
          onReceiveProgress: onReceiveProgress);
      res.data = decoder?.call(res.data) ?? res.data;
      return res as Response<T>;
    } catch (error) {
      int errorCode = 0;
      RequestOptions requestOptions = RequestOptions(path: path);
      if (error is DioError) {
        requestOptions = error.requestOptions;
        final errorTypes = DioErrorType.values;
        for (int i = 0; i < errorTypes.length; i++) {
          if (errorTypes[i].index == error.type.index) {
            errorCode = i;
          }
        }
      }
      return Response(
        requestOptions: requestOptions,
        statusCode: errorCode * -1,
      );
    }
  }

  /// check Options.
  static Options _checkOptions(Method method, Options? options) {
    if (options == null) {
      options = new Options();
    }
    options.method = method.method;
    return options;
  }

  /// mock的数据解析，
  static Future<Response<T>> _mock<T>(
    PQNetworkDataDecoder<T>? decoder,
    dynamic mockData,
    String? url,
    String method,
  ) {
    assert(mockData != null, "mock的时候，mockData不能传空");
    try {
      T res;

      /// 实现了decoder，就使用decoder
      if (decoder != null) {
        res = decoder(mockData!);
      } else {
        /// dynamic
        res = mockData;
      }

      return Future.value(
        Response(
          data: res,
          statusCode: 200,
          statusMessage: "Mock Data Success",
          requestOptions:
              RequestOptions(path: url ?? 'null path', method: method),
        ),
      );
    } catch (e) {
      return Future.value(
        Response(
          statusCode: 0,
          statusMessage: "Mock Data Failed",
          requestOptions:
              RequestOptions(path: url ?? 'null path', method: method),
        ),
      );
    }
  }
}
